@@ -182,7 +182,8 @@ def _attribute_transformer(prefix: str, scrub_message: str, keep: bool) -> Calla
     """
     Create a function which may be used to transform exception attributes.
 
-    If an attribute is string-valued, apply the logic keep? prefix + attr: prefix
+    If an attribute is string-valued or int-valued,
+    apply the logic keep? prefix + attr: prefix
     + scrub_message.
 
     If the attribute is iterable, apply this logic to each member of the
@@ -200,6 +201,9 @@ def _attribute_transformer(prefix: str, scrub_message: str, keep: bool) -> Calla
                 rv = prefix + o
             else:
                 rv = prefix + scrub_message
+        elif isinstance(o, int):
+            # handling SystemExit
+            rv = inner(str(o))
         elif isinstance(o, Iterable):
             rv = type(o)(map(inner, o))  # type: ignore
         elif callable(o):
