@@ -4,8 +4,8 @@
 """
 Pipeline helper class to create pipelines loading modules from a flexible manifest.
 """
+from __future__ import annotations
 import argparse
-from collections import defaultdict
 import os
 import sys
 import json
@@ -16,8 +16,8 @@ import uuid
 import shutil
 import yaml
 import jsonpath_ng
-from typing import Callable, Optional, Set
-from toposort import toposort_flatten
+from typing import Any, Callable, Dict, Hashable, List, Optional, Set, Tuple, Union
+from toposort import toposort_flatten, CircularDependencyError
 import datetime
 from dataclasses import dataclass
 from functools import lru_cache, wraps
@@ -37,6 +37,7 @@ try:
     from azureml.pipeline.core import PipelineRun
     from azure.ml.component.component import Component, Input, Output
     from azure.ml.component.pipeline import Pipeline
+    from azure.ml.component._pipeline_parameters import PipelineParameter
     from azureml.exceptions._azureml_exception import UserErrorException
 except ImportError as error:
     raise ImportError(
@@ -1726,7 +1727,9 @@ class AMLPipelineHelper:
             except BaseException as e:
                 log.error(f"An error occurred, recovery is not successful: {e}")
 
-    def _update_dc_configs(self, pipeline):
+    def _update_dc_configs(self, pipeline: Pipeline, debug: bool = False) -> Pipeline:
+
+        # collect the known DC targets
         # workaround for missing "throw_on_missing" in currently used omegaconf
         compute_config_as_dict = yaml.safe_load(OmegaConf.to_yaml(self.config.compute))
         dc_targets: Set[str] = {
@@ -1735,98 +1738,93 @@ class AMLPipelineHelper:
             if key.endswith("_dc_target") and target and target != str(MISSING)
         }
         if not dc_targets:
-            log.info("The compute config does not involve any DC targets.")
+            # This is kind of pointless because if one local component exists
+            # then one of the `apply_*` methods wil fail if no dc target is not
+            # specified. And if no local component exists, then there is no
+            # point in outputting this. But let's do it anyway:
+            log.info("The compute config does not contain any DC targets.")
             return pipeline
-        nodes = pipeline._expand_pipeline_nodes()
-        graph = defaultdict(list)
-        node_mapping = {}
-        for n in nodes:
-            node_mapping[n._instance_id] = n
-            for input in n.inputs.values():
-                try:
-                    owner = (
-                        input._get_data_owner()
-                    )  # points to subgraph node if upstream component is subgraph
-                    if owner:
-                        graph[n._instance_id].append(owner._instance_id)
 
-                except AttributeError:
-                    continue
-        subpipelines = pipeline._expand_pipeline_to_pipelines()
-        for n in subpipelines:
-            # https://msdata.visualstudio.com/DefaultCollection/Vienna/_workitems/edit/2026740/
-            # if the last node of a subgraph needs to be updated
-            # then we need to update subgraph.outputs datastore too
-            node_mapping[n._instance_id] = n
-            graph[n._instance_id].append(n.nodes[-1]._instance_id)
-        sorted_node_ids = toposort_flatten(graph)
-        use_dc = defaultdict(bool)
-        for id in sorted_node_ids:
-            if id not in node_mapping:
+        log.info("Overriding compute target if upstream steps use DC...")
+
+        # build and parse the dependency graph
+        dependency_graph = _DependencyGraph.from_pipeline(
+            pipeline=pipeline, debug=debug
+        )
+
+        should_use_dc = dependency_graph.determine_dc_nodes(dc_targets=dc_targets)
+
+        log.info(f"Need to configure the following for DC: {sorted(should_use_dc)}")
+
+        # prepare a mapping from known prod targets to (their corresponding
+        # key in the shrike config and) the corresponding DC target
+        target_map: Dict[str, Tuple[str, Optional[str]]] = {}
+        for prod_target in [
+            "hdi_prod_target",
+            "linux_cpu_prod_target",
+            "linux_gpu_prod_target",
+            "windows_cpu_prod_target",
+            "synapse_prod_target",
+        ]:
+            if prod_target not in self.config.compute:
                 continue
-            node = node_mapping[id]
-            # scope component doesn't configure compute target
-            if node.type != "ScopeComponent":
-                if not isinstance(node.runsettings.target, str):
-                    log.warn(
-                        f"The compute target of node {node.name} is a PipelineParameter. Shrike will not update even if this or the upstream steps use HDI DC or Synapse DC. Please update manually."
-                    )
-                elif node.runsettings.target in dc_targets:
-                    use_dc[id] = True
-                    continue
-            for parent in graph[id]:
-                if use_dc[parent]:
-                    use_dc[id] = True
-                    break
+            target_map[self.config.compute[prod_target]] = [
+                prod_target,
+                self.config.compute.get(prod_target.replace("_prod_", "_dc_"), None),
+            ]
 
-        for id, node in node_mapping.items():
-            if use_dc[id]:
-                if node.type == "ScopeComponent":
-                    # scope component doesn't need to configure compute target
+        # move to DC!
+        for node_id in should_use_dc:
+            node = dependency_graph.nodes[node_id]
+
+            if isinstance(node, Output):
+
+                if node._owner.type == "ScopeComponent":
                     # output datastore has to be cosmos (ADLSg1), while dc_datastore is ADLSg2
                     continue
-                self._set_all_outputs_to(
-                    node,
+
+                dc_datastore = self.config.compute.dc_datastore
+                node.configure(
+                    datastore=dc_datastore,
                     output_mode=None,
-                    datastore_name=self.config.compute.dc_datastore,
                 )
-                log.info(
-                    f"Writing outputs of {node.name} to dc_datastore {self.config.compute.dc_datastore}"
+                log.info(f"Configured {node_id} to use datastore {dc_datastore}.")
+
+            elif not hasattr(node.runsettings, "target"):
+                # e.g. scope component doesn't need to configure compute target
+                log.info(f"Node {node.name} does not have `target`.")
+                continue
+            elif not isinstance(node.runsettings.target, str):
+                log.warn(
+                    f"The compute target of node {node.name} is a PipelineParameter. Shrike will not update even if this or the upstream steps use DC. Please update manually."
                 )
-                if not isinstance(node.runsettings.target, str):
-                    log.info(
-                        f"Node {node.name}'s target is a PipelineParameter, skipping."
+                continue
+            elif node.runsettings.target in dc_targets:
+                # nothing to do here
+                pass
+
+            elif node.runsettings.target not in target_map:
+                log.error(
+                    f"Node {node} is using an unknown target: "
+                    f"{node.runsettings.target}. "
+                    f"Please move this to DC manually, or re-run "
+                    f"this after adding the target to the compute config."
+                )
+
+            else:
+                prod_target_key, dc_target = target_map[node.runsettings.target]
+                if dc_target is None:
+                    log.error(
+                        f"No corresponding DC target is defined for "
+                        f"prod target {prod_target_key}. "
+                        f"Please move this to DC manually, or re-run "
+                        f"this after adding the DC target to the compute config."
                     )
                     continue
-                elif node.runsettings.target == self.config.compute.hdi_prod_target:
-                    node.runsettings.configure(target=self.config.compute.hdi_dc_target)
-                elif (
-                    node.runsettings.target == self.config.compute.linux_cpu_prod_target
-                ):
-                    node.runsettings.configure(
-                        target=self.config.compute.linux_cpu_dc_target
-                    )
-                elif (
-                    node.runsettings.target == self.config.compute.linux_gpu_prod_target
-                ):
-                    node.runsettings.configure(
-                        target=self.config.compute.linux_gpu_dc_target
-                    )
-                elif (
-                    node.runsettings.target
-                    == self.config.compute.windows_cpu_prod_target
-                ):
-                    node.runsettings.configure(
-                        target=self.config.compute.windows_cpu_dc_target
-                    )
-                elif node.runsettings.target == self.config.compute.synapse_prod_target:
-                    node.runsettings.configure(
-                        target=self.config.compute.synapse_dc_target
-                    )
-                else:
-                    continue
+
+                node.runsettings.configure(target=dc_target)
                 log.info(
-                    f"Updating compute target for {node.name} {id} to {node.runsettings.target}"
+                    f"Updating compute target for {node_id} to {node.runsettings.target}"
                 )
         return pipeline
 
@@ -2158,3 +2156,144 @@ Follow link below to access your pipeline run directly:
         hydra_run()
 
         return cls.BUILT_PIPELINE  # return so we can have some unit tests done
+
+
+class _DependencyGraph:
+    NodeType = Union[Component, Output]
+
+    graph: Dict[str, List[str]] = {}
+    nodes: Dict[str, NodeType] = {}
+    debug: bool = False
+
+    def map_to_id(self, item: Union[Pipeline, Component, Output]) -> Optional[str]:
+        """Maps the given item to an id."""
+        # Reminder: class-wise, pipelines are also components
+        extra = f"{item.comment}, " if self.debug and hasattr(item, "comment") else ""
+        if isinstance(item, Pipeline):
+            return f"pipeline `{item.display_name.replace(' ', '-')}` ({extra}{item._instance_id})"
+        if isinstance(item, Component):
+            return f"component `{item.display_name.replace(' ', '-')}` ({extra}{item._instance_id})"
+        if isinstance(item, Output):
+            return f"output `{item.port_name}` of {self.map_to_id(item._owner)}"
+        raise ValueError(
+            f"Please report this. Could not map {item} to dependency graph id."
+        )
+
+    def add_edge(
+        self,
+        from_node: NodeType,
+        to_node: NodeType,
+    ) -> None:
+        """Add an edge from a dependent node its dependency."""
+        for node in [from_node, to_node]:
+            if not isinstance(node, (Component, Output)):
+                raise ValueError(
+                    f"Please report this. "
+                    f"Expected a valid node type but was given {node} as node."
+                )
+            node_id = self.map_to_id(node)
+            known_node = self.nodes.get(node_id)
+            if known_node is not None and node is not known_node:
+                raise ValueError("Report this. Two different nodes with same id.")
+            self.nodes[node_id] = node
+
+        ids = [self.map_to_id(node) for node in [from_node, to_node]]
+        if ids[0] == ids[1]:
+            return
+        to_nodes = self.graph.setdefault(ids[0], [])
+        if ids[1] not in to_nodes:  # yeah it's a list but short
+            to_nodes.append(ids[1])
+        self.assert_acyclic()
+        log.debug(f"Add edge {ids})")
+
+    def topsorted_ids(self) -> List[str]:
+        return toposort_flatten(self.graph)
+
+    def assert_acyclic(self) -> None:
+        try:
+            self.topsorted_ids()
+        except CircularDependencyError as e:
+            raise ValueError(
+                "Error while constructing dependency graph for DC management."
+            ) from e
+
+    @classmethod
+    def from_pipeline(cls, pipeline: Pipeline, debug: bool) -> _DependencyGraph:
+        graph = _DependencyGraph()
+        graph.debug = debug
+
+        def resolve(node: Any) -> Optional[_DependencyGraph.NodeType]:
+            if isinstance(node, Output) or isinstance(node, Component):
+                return node
+            if isinstance(node, Input):
+                input_dset = node._dset
+                if input_dset is None:
+                    return None
+                return resolve(input_dset)
+            if isinstance(node, PipelineParameter):
+                default_value = node.default_value
+                if default_value is None or default_value is node:
+                    # can't resolve this
+                    return None
+                return resolve(default_value)
+            return None
+
+        for subpipeline in pipeline._expand_pipeline_to_pipelines():
+            # For DC, we only need to adapt `Output` and `Component`
+            # components. The `Output`s of pipelines do not coincide
+            # with the `Output`s of components, hence we have to
+            # insert those into our dependency graph:
+            for output_name, output in subpipeline.outputs.items():
+                linked_output = subpipeline._outputs_mapping[output_name]
+                graph.add_edge(output, linked_output)
+            # However, it seems we can skip the pipeline inputs.
+
+        for component in pipeline._expand_pipeline_nodes():
+            for output in component.outputs.values():
+                graph.add_edge(output, component)
+            for input_ in component.inputs.values():
+                corresponding_node = resolve(input_)
+                if corresponding_node is not None:
+                    graph.add_edge(component, corresponding_node)
+        return graph
+
+    def determine_dc_nodes(self, dc_targets: List[str]) -> Set[str]:
+        """Determine which downstream components/outputs should be configured for DC.
+
+        Args:
+            dc_targets: known DC targets (used to determine which nodes
+               are already configured for DC)
+        Returns:
+            list of node ids for all nodes that should be configured for
+            dc (or have already been configured for dc)
+        """
+        should_use_dc: Set[str] = set()
+        for node_id in self.topsorted_ids():
+
+            # use dc, if upstream used dc
+            for dependency in self.graph.get(node_id, []):
+                if dependency in should_use_dc:
+                    should_use_dc.add(node_id)
+                    continue
+
+            item = self.nodes[node_id]
+
+            if not isinstance(item, Component):
+                # If users choose to move an output to iso datastore out of
+                # nowhere, we ignore this. We're only caring about cases where
+                # the (`apply_*` method caused the) run settings of components
+                # to be configured for DC - and fix downstream components.
+                continue
+
+            if not hasattr(item.runsettings, "target") or not isinstance(
+                item.runsettings.target, str
+            ):
+                # e.g. Scope components do not configure compute targets
+                # e.g. if `target` is a PipelineParameter, it returns None
+                # and no need to update
+                continue
+
+            if item.runsettings.target in dc_targets:
+                should_use_dc.add(node_id)
+
+        return should_use_dc
